from collections.abc import Callable, Hashable, Iterable
from functools import reduce
from typing import TypeVar

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)
TPost = TypeVar("TPost")
TPre = TypeVar("TPre")
K = TypeVar("K", bound=Hashable)
V = TypeVar("V")
type Filterer[T] = Callable[[T], bool]

type IterableVal = list["ValType"] | set[Hashable] | tuple["ValType", ...]
type ValType = Hashable | IterableVal | dict[Hashable, "ValType"]
type IterReturnType = list[ValType] | set[Hashable] | tuple[ValType]
type Mergeable[K, T] = str | object | list[T] | set[T] | tuple[T] | dict[K, "Mergeable"]


def make_hashable(ob: object) -> Hashable:
    def make_sequence_hashable(ob: list[T] | set[T]) -> tuple[T, ...]:
        return tuple(ob)

    def make_dict_hashable(ob: dict[K, V]) -> tuple[tuple[K, V], ...]:
        return tuple(ob.items())

    if isinstance(ob, Hashable):
        return ob
    if isinstance(ob, list | set):
        return make_sequence_hashable(ob)
    if isinstance(ob, dict):
        return make_dict_hashable(ob)
    return str(ob)


def join_as_sequence(  # noqa: C901,PLR0911
    lv: ValType,
    rv: ValType,
    default_new_type: type[list] | type[tuple] | type[set] = tuple,
    add_duplicates: bool = True,
) -> IterableVal:
    def keep_duplicates(a: Iterable, _) -> Iterable:
        return a

    def remove_duplicates(a: Iterable, b) -> Iterable:
        compare = set(b) if isinstance(b, list | set | tuple) else {b}
        return filter(lambda x: x not in compare, a)

    filtr = keep_duplicates if add_duplicates else remove_duplicates

    if isinstance(lv, list):
        if isinstance(rv, list | set | tuple):
            return lv + list(filtr(rv, lv))
        else:
            return lv if rv in lv else [*lv, rv]
    if isinstance(lv, tuple):
        if isinstance(rv, list | set | tuple):
            return (*lv, *filtr(rv, lv))
        else:
            return lv if rv in lv else (*lv, rv)
    if isinstance(lv, set):
        if isinstance(rv, list | set | tuple):
            return {*lv, *filtr(map(make_hashable, rv), lv)}
        else:
            return lv if make_hashable(rv) in lv else {*lv, make_hashable(rv)}

    if isinstance(rv, list):
        return [lv, *filtr(rv, lv)]
    if isinstance(rv, tuple):
        return (lv, *filtr(rv, lv))
    if isinstance(rv, set):
        return {make_hashable(lv), *filtr(rv, make_hashable(lv))}

    return default_new_type((lv, rv))


def merge_dicts(
    *dicts: dict[Hashable, ValType],
    default_new_sequence: type[list] | type[tuple] | type[set] = tuple,
    add_duplicates: bool = True,
) -> dict[Hashable, ValType]:
    def merge_values(right_val: ValType, left_val: ValType) -> ValType:  # noqa: PLR0911
        if left_val == right_val:
            return left_val
        if isinstance(left_val, dict) and isinstance(right_val, dict):
            return merge_dicts(left_val, right_val, default_new_sequence=default_new_sequence)
        if left_val is None:
            return right_val
        if right_val is None:
            return left_val
        if isinstance(left_val, dict):
            return left_val
        if isinstance(right_val, dict):
            return right_val
        return join_as_sequence(
            left_val,
            right_val,
            default_new_type=default_new_sequence,
            add_duplicates=add_duplicates,
        )

    def merge(
        left: dict[Hashable, ValType], right: dict[Hashable, ValType]
    ) -> dict[Hashable, ValType]:
        common_keys: set[Hashable] = set(left).intersection(set(right))
        common: dict[Hashable, ValType] = {}
        for key in common_keys:
            lv, rv = left[key], right[key]
            new_dict: dict[Hashable, ValType] = {key: merge_values(lv, rv)}
            common.update(new_dict)

        return left | right | common

    return reduce(merge, dicts)



from datetime import datetime
from pathlib import Path


def make_timestamp(places: int = 3) -> str:
    dt = datetime.now()
    if places > 0:
        idx: int | None = min(0, int(places) - 6) or None
        return dt.strftime("%Y-%m-%d_%H:%M:%S.%f")[:idx]
    return dt.strftime("%Y-%m-%d_%H:%M:%S")


def insert_timestamp(p: Path | str) -> Path:
    new_path = str(p)

    if "." in new_path:
        base, suffix = new_path.rsplit(".", 1)
        return Path(f"{base}__{make_timestamp()}.{suffix}")

    return Path(f"{new_path}__{make_timestamp()}")



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Self

from .pydantic_extensions import BaseModelRW

# class PathsManager:
#     @abstractmethod
#     def setup(self) -> None: ...

#     @classmethod
#     @abstractmethod
#     def auto(cls, root_dir: Path): ...

#     @classmethod
#     @abstractmethod
#     def read(cls, config_file_path: Path) -> Self: ...

#     @abstractmethod
#     def write(self, config_file_path: Path) -> None: ...


class PathsManager(BaseModelRW, ABC):
    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs) -> Self:
        """
        Create directories first if they do not already exist.
        """
        ...

    @classmethod
    @abstractmethod
    def auto(cls, *args, **kwargs) -> Self:
        """
        Create directories and files in the default location if they do not alreay exist.
        """
        ...

    @classmethod
    @abstractmethod
    def read(cls, config_file_path: Path) -> Self: ...

    @abstractmethod
    def write(self, config_file_path: Path) -> None: ...

    @staticmethod
    def ensure_file_exists(p: Path, content_if_empty: str = "") -> Path:
        if not p.exists():
            p.touch()
            if content_if_empty:
                p.write_text(content_if_empty)
        return p

    @staticmethod
    def ensure_directory_exists(p: Path) -> Path:
        for parent in p.parents:
            if not parent.exists():
                parent.mkdir()
        if not p.exists():
            p.mkdir()
        return p



import re
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable
from dataclasses import dataclass
from pathlib import Path
from types import UnionType
from typing import Annotated, Any, Protocol, Self, TypeVar

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)


type Atomic = int | float | str | bool | None
type ClassInfo = type | UnionType | tuple["ClassInfo"]
type JSONPrimitive = int | float | str | None
type JSONDict = dict[str, JSONPrimitive | "JSONDict" | list["JSONDict | JSONPrimitive"]]
# alternative: JSONDict = dict[str, Union["JSONDict", str, list[str], tuple[str, ...], set[str]]]
type JSONList = list[JSONPrimitive | JSONDict]
type Endofunction[T] = Callable[[T], T]
type SequenceAlias[T] = list[T] | tuple[T] | set[T] | map[T] | filter[T]
_SequenceAliasVar = list | tuple | set | map | filter


def areinstances(iterable_instance: Iterable[Any], class_or_tuple: ClassInfo) -> bool:
    return all(map(lambda inst: isinstance(inst, class_or_tuple), iterable_instance))


def fallback_if_none(orig: T | None, alt: T) -> T:
    return alt if (orig is None) else orig


def call_fallback_if_none(orig: T | None, alt: Callable[[], T]) -> T:
    return alt() if (orig is None) else orig


class SupportsGe(Protocol):
    def __ge__(self: T, __other: T) -> bool: ...


class SupportsGt(Protocol):
    def __gt__(self: T, __other: T) -> bool: ...


class SupportsLe(Protocol):
    def __le__(self: T, __other: T) -> bool: ...


class SupportsLt(Protocol):
    def __lt__(self: T, __other: T) -> bool: ...


@dataclass(frozen=True)
class _BaseMetadata:
    """Base class for all metadata.

    This exists mainly so that implementers
    can do `isinstance(..., BaseMetadata)` while traversing field annotations.
    """

    __slots__ = ()


@dataclass(frozen=True)
class Ge(_BaseMetadata):
    """Ge(ge=x) implies that the value must be greater than or equal to x.

    It can be used with any type that supports the ``>=`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    ge: SupportsGe


@dataclass(frozen=True)
class Gt(_BaseMetadata):
    """Gt(gt=x) implies that the value must be greater than x.

    It can be used with any type that supports the ``>`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    gt: SupportsGt


@dataclass(frozen=True)
class Lt(_BaseMetadata):
    """Lt(lt=x) implies that the value must be less than x.

    It can be used with any type that supports the ``<`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    lt: SupportsLt


@dataclass(frozen=True)
class Le(_BaseMetadata):
    """Le(le=x) implies that the value must be less than x.

    It can be used with any type that supports the ``<=`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    le: SupportsLe


class SupportsIO(ABC):
    @classmethod
    @abstractmethod
    def read(cls, read_path: Path) -> Self: ...

    @abstractmethod
    def write(self, write_path: Path) -> None: ...


class Pattern:
    DATE: re.Pattern[str] = re.compile(r"^[12]\d\d\d-(0?\d|1[012]|)-(0?\d|[12]\d|3[01])$")
    DATE_STRICT: re.Pattern[str] = re.compile(r"^[12]\d\d\d-(0\d|1[012]|)-(0\d|[12]\d|3[01])$")
    DATE_LOOSE: re.Pattern[str] = re.compile(r"(\d{2,4})[^\d](\d\d?)[^\d](\d\d?)")
    ID: re.Pattern[str] = re.compile(r"^[A-Za-z][A-Za-z0-9_]+$")
    ID_OR_2IDS: re.Pattern[str] = re.compile(r"^[A-Za-z][A-Za-z0-9_]+,[A-Za-z][A-Za-z0-9_]+$")
    NATURAL: re.Pattern[str] = re.compile(r"^[1-9][0-9]*$")
    PROPORTION: re.Pattern[str] = re.compile(r"^0?\.[0-9]+$")
    PROPORTION_OR_2: re.Pattern[str] = re.compile(r"^0?\.[0-9]+$|^0?\.[0-9]+,0?\.[0-9]+$")


TimeAmountRaw = str | int

Natural = Annotated[int, Ge(ge=0)]
Nonnegative = Annotated[float, Ge(ge=0)]
Positive = Annotated[float, Gt(gt=0)]
PositiveScore = Annotated[float, Gt(gt=0.0), Lt(lt=5.0)]
NegativeScore = Annotated[float, Gt(gt=-5.0), Lt(lt=0.0)]
Proportion = Annotated[float, Ge(ge=0.0), Le(le=1.0)]
NonnegativeInt = Annotated[int, Ge(ge=0)]
NonnegativeFloat = Annotated[float, Ge(ge=0.0)]
PolarityScore = Annotated[float, Ge(ge=-1.0), Le(le=1.0)]


def ensure_set(x: H | SequenceAlias[H]) -> set[H]:
    return set(x) if isinstance(x, _SequenceAliasVar) else {x}


def ensure_tuple(x: T | SequenceAlias[T]) -> tuple[T, ...]:
    return tuple(x) if isinstance(x, _SequenceAliasVar) else (x,)


def ensure_list(x: T | SequenceAlias[T]) -> list[T]:
    return list(x) if isinstance(x, _SequenceAliasVar) else [x]



from collections.abc import Callable
from operator import not_ as negate


def nearly_equal(a: int | float, b: int | float) -> bool:
    return abs(a - b) < 0.00001


def equal_within(a: int | float, b: int | float, epsilon: float | int) -> bool:
    return abs(a - b) < epsilon


def match_empty(item: str | None | bool | set[str] | float) -> bool:
    return not bool(item)


def trivial_condition(n: object) -> bool:
    return True


def as_bool(object_: object) -> bool:
    return bool(object_)


def make_maybe_not(_negated: bool) -> Callable[[object], bool]:
    return negate if _negated else as_bool



import json
import os
import shutil
from pathlib import Path
from typing import Literal

from .functional import tmap
from .timestamping import make_timestamp
from .typing_utils import JSONDict, JSONList


def ensure_path(p: Path | str) -> Path:
    if isinstance(p, Path):
        return p
    return Path(p)


def list_full(directory: str | Path, ending: str = "") -> list[Path]:
    directory = ensure_path(directory)
    return sorted([directory / file for file in os.listdir(directory) if file.endswith(ending)])


def read_raw(json_path: Path) -> str:
    with open(json_path, encoding="utf-8") as f:
        return f.read()


def back_up_json(original_path: Path, backup_dir: Path | None) -> None:
    if not backup_dir:
        return
    backup_path = backup_dir / original_path.name.replace(".json", f"__{make_timestamp()}.json")
    if original_path.exists():
        shutil.copy(original_path, backup_path)


def read_json(json_path: Path, backup_dir: Path | None = None) -> JSONDict | JSONList:
    back_up_json(json_path, backup_dir)
    with open(json_path, encoding="utf-8") as f:
        return json.load(f)


def write_json(
    python_obj: JSONDict | JSONList | bytes,
    json_path: Path,
    backup_dir: Path | None = None,
    mode: Literal["w", "a"] = "w",
) -> None:
    if mode not in {"w", "a"}:
        raise ValueError(f"Invalid mode: '{mode}'")
    back_up_json(json_path, backup_dir)
    if mode == "a":
        existing = read_json(json_path)
        if isinstance(existing, list) and isinstance(python_obj, list):
            python_obj = existing + python_obj
        elif isinstance(existing, dict) and isinstance(python_obj, dict):
            python_obj = existing | python_obj
        else:
            raise ValueError("Data types do not match.")
    with open(json_path, "w", encoding="utf-8") as f:
        json.dump(python_obj, f, ensure_ascii=False, indent=4)


def write_raw(text: str, file_path: Path) -> None:
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(text)


def write_raw_bytes(text: bytes, json_path: Path) -> None:
    with open(json_path, "wb") as f:
        f.write(text)


def list_names(directory: Path, ending: str) -> tuple[str, ...]:
    return tmap(
        lambda p: p.name.replace(ending, ""),
        list_full(directory, ending=ending),
    )


def name_from_full(p: Path | str, ending: str = ".json") -> str:
    if isinstance(p, Path):
        return p.name.replace(ending, "")
    return p.split("/")[-1].replace(ending, "")


def remove_all_under(dir: Path) -> None:
    for item in dir.iterdir():
        if item.is_file() or item.is_symlink():
            item.unlink()
        elif item.is_dir():
            shutil.rmtree(item)



from collections.abc import Callable, Hashable, Iterable
from functools import reduce
from typing import TypeVar, overload

from multipledispatch import dispatch

T = TypeVar("T")
Seq = TypeVar("Seq", bound=list | tuple)
TPost = TypeVar("TPost")
TPre = TypeVar("TPre")
K = TypeVar("K", bound=Hashable)
V = TypeVar("V")
type Filterer[T] = Callable[[T], bool]


@overload
def endomap(callable_: Callable[[TPre], TPost], sequence: list[TPre]) -> list[TPost]: ...
@overload
def endomap(callable_: Callable[[TPre], TPost], sequence: set[TPre]) -> set[TPost]: ...
@overload
def endomap(
    callable_: Callable[[TPre], TPost], sequence: tuple[TPre, ...]
) -> tuple[TPost, ...]: ...


def endomap(callable_, sequence):
    return type(sequence)(map(callable_, sequence))


@overload
def endofilter(callable_: Callable[[T], bool], sequence: list[T]) -> list[T]: ...
@overload
def endofilter(callable_: Callable[[T], bool], sequence: set[T]) -> set[T]: ...
@overload
def endofilter(callable_: Callable[[T], bool], sequence: tuple[T, ...]) -> tuple[T, ...]: ...


def endofilter(callable_, sequence):
    return type(sequence)(filter(callable_, sequence))


def lmap(callable_: Callable[[TPre], TPost], iterable: Iterable[TPre]) -> list[TPost]:
    return list(map(callable_, iterable))


def smap(callable_: Callable[[TPre], TPost], iterable: Iterable[TPre]) -> set[TPost]:
    return set(map(callable_, iterable))


def tmap(callable_: Callable[[TPre], TPost], iterable: Iterable[TPre]) -> tuple[TPost, ...]:
    return tuple(map(callable_, iterable))


def vmap(callable_: Callable[[TPre], TPost], dictionary: dict[K, TPre]) -> dict[K, TPost]:
    return {k: callable_(v) for k, v in dictionary.items()}


def kmap(callable_: Callable[[TPre], TPost], dictionary: dict[TPre, V]) -> dict[TPost, V]:
    return {callable_(k): v for k, v in dictionary.items()}


def dmap(callable_: Callable[[TPre], TPost], dictionary: dict[TPre, TPre]) -> dict[TPost, TPost]:
    return {callable_(k): callable_(v) for k, v in dictionary.items()}


def lfilter(filterer: Filterer[T], iterable: Iterable[T]) -> list[T]:
    return list(filter(filterer, iterable))


def sfilter(filterer: Filterer[T], iterable: Iterable[T]) -> set[T]:
    return set(filter(filterer, iterable))


def tfilter(filterer: Filterer[T], iterable: Iterable[T]) -> tuple[T, ...]:
    return tuple(filter(filterer, iterable))


def vfilter(filterer: Filterer[V], dictionary: dict[K, V]) -> dict[K, V]:
    return {k: v for k, v in dictionary.items() if filterer(v)}


def kfilter(filterer: Filterer[K], dictionary: dict[K, V]) -> dict[K, V]:
    return {k: v for k, v in dictionary.items() if filterer(k)}


def dfilter(filterer: Filterer[T], dictionary: dict[T, T]) -> dict[T, T]:
    return {k: v for k, v in dictionary.items() if filterer(k) and filterer(v)}


def identity[T](x: T) -> T:
    return x


def fold_dictionaries[K, V](dicts: Iterable[dict[K, V]]) -> dict[K, V]:
    def _or(dict1: dict[K, V], dict2: dict[K, V]) -> dict[K, V]:
        return dict1 | dict2

    return reduce(_or, dicts)


@dispatch(tuple)
def split_head[T](seq: tuple[T, ...]) -> tuple[T, tuple[T, ...]]:
    if not seq:
        raise ValueError(f"Tuple {seq} is invalid because it cannot be properly split.")
    return seq[0], seq[1:]


@dispatch(list)
def split_head[T](seq: list[T]) -> tuple[T, list[T]]:
    if not seq:
        raise ValueError(f"List {seq} is invalid because it cannot be properly split.")
    return seq[0], seq[1:]


@dispatch(list)
def tail[T](seq: list[T]) -> list[T]:
    return seq[1:]


@dispatch(tuple)
def tail[T](seq: tuple[T, ...]) -> tuple[T, ...]:
    return seq[1:]



import re

from .typing_utils import areinstances


def re_bool_search(p: re.Pattern[str], s: str) -> bool:
    return bool(re.search(p, s))


def re_split(expr: str | re.Pattern[str], s: str) -> list[str]:
    segments = re.split(expr, s)
    if not areinstances(segments, str):
        raise TypeError("Function 're_split' may only return a list of strings.")
    return segments


def escape_character(s: str) -> str:
    return {
        ".": "\\.",
        "?": "\\?",
        "|": "\\|",
        "*": "\\*",
        "+": "\\+",
        "^": "\\^",
    }.get(s, s)



"""
Idea: make BaseSequence[C, T] type,
    where C is list|set|tuple and T is the element type
"""

import json
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable, Mapping
from pathlib import Path
from types import GenericAlias
from typing import (
    Any,
    Literal,
    Protocol,
    Self,
    TypeVar,
    cast,
    get_args,
    runtime_checkable,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    TypeAdapter,
)
from pydantic.config import ExtraValues
from pydantic.fields import FieldInfo
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    JsonSchemaMode,
    JsonSchemaValue,
)
from pydantic.main import IncEx
from pydantic_core import CoreSchema, core_schema

from .io_utils import JSONDict, read_json, write_json

T = TypeVar("T")
K_ = TypeVar("K_", bound=Hashable)
V_ = TypeVar("V_")
Mode = Literal["json", "python"] | str


VALID_MODES: set[Literal["json", "python"]] = {"json", "python"}


@runtime_checkable
class ComplexValidatedProtocol(Protocol):
    """
    Protocol defining minimum shared behavior between AbstractCustom, its children, and BaseModelRW.
    """

    @classmethod
    def read_json_file(cls, source: Path) -> Self: ...

    def write_json_file(self, destination: Path) -> Path: ...


class BaseModelRW(BaseModel):
    @classmethod
    def read_json_file(cls, source: Path) -> Self:
        return cls.model_validate(read_json(source))

    def write_json_file(self, destination: Path) -> Path:
        write_json(self.model_dump(mode="json"), destination)
        return destination


class AbstractCustom(ABC):
    model_config: ConfigDict = ConfigDict()
    __pydantic_core_schema__: CoreSchema

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.model_validate,
            core_schema.dict_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        return json_schema

    def __init__(self, *args, **kwargs) -> None:
        raise TypeError(f"Use `model_validate` to instantiate {self.__class__.__name__}")

    @classmethod
    @abstractmethod
    def get_adapter(cls) -> TypeAdapter: ...

    @abstractmethod
    def __repr__(self) -> str: ...

    @abstractmethod
    def __str__(self) -> str: ...

    @classmethod
    def model_fields(cls) -> dict[str, FieldInfo]:
        return {}

    @property
    def model_extra(self) -> dict[str, Any] | None:
        return None

    @property
    def model_fields_set(self) -> set[str]:
        return set()

    @classmethod
    def model_construct(cls, _fields_set: set[str] | None = None, **values: Any) -> Self:
        raise NotImplementedError

    def model_copy(self, *, update: Mapping[str, Any] | None = None, deep: bool = False) -> Self:
        return self.__class__.model_validate(self.model_dump())

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_computed_fields: bool = False,
        round_trip: bool = False,
        warnings: (bool | Literal["none", "warn", "error"]) = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        if mode not in VALID_MODES:
            raise ValueError
        if exclude_computed_fields:
            raise ValueError

        return self.get_adapter().dump_python(
            self.pre_dump_hook(mode=mode),
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )

    def model_dump_json(
        self,
        *,
        indent: int | None = None,
        ensure_ascii: bool = False,
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_computed_fields: bool = False,
        round_trip: bool = False,
        warnings: (bool | Literal["none", "warn", "error"]) = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> str:
        if ensure_ascii or exclude_computed_fields:
            raise ValueError
        from_hook = self.pre_dump_hook(mode="json")
        return (
            self.get_adapter()
            .dump_json(
                from_hook,
                indent=indent,
                include=include,
                exclude=exclude,
                context=context,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
                fallback=fallback,
                serialize_as_any=serialize_as_any,
            )
            .decode("utf-8")
        )

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[GenerateJsonSchema] = GenerateJsonSchema,
        mode: JsonSchemaMode = "validation",
        *,
        union_format: Literal["any_of", "primitive_type_array"] = "any_of",
    ) -> dict[str, Any]:
        return cls.get_adapter().json_schema(
            by_alias=by_alias,
            ref_template=ref_template,
            schema_generator=schema_generator,
            mode=mode,
        )

    @classmethod
    def model_parametrized_name(cls, params: tuple[type[Any], ...]) -> str:
        raise NotImplementedError

    @classmethod
    def model_validate(
        cls,
        obj: Any,
        *,
        strict: bool | None = None,
        extra: ExtraValues | None = None,
        from_attributes: bool | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
    ) -> Self:
        if extra:
            raise ValueError
        adapter = cls.get_adapter()
        obj = cls.pre_validation_hook_python(obj)
        obj = adapter.validate_python(
            obj,
            strict=strict,
            from_attributes=from_attributes,
            context=context,
            by_alias=by_alias,
            by_name=by_name,
        )
        validated = cls(obj)
        return cls.post_validation_hook(validated)

    @classmethod
    def model_validate_json(
        cls,
        json_data: str | bytes | bytearray,
        *,
        strict: bool | None = None,
        extra: ExtraValues | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
    ) -> Self:
        if extra:
            raise ValueError
        adapter = cls.get_adapter()
        json_string = cls.pre_validation_hook_json(json_data)
        raw_validated = adapter.validate_json(
            json_string,
            strict=strict,
            context=context,
            by_alias=by_alias,
            by_name=by_name,
        )
        validated = cls(raw_validated)
        return cls.post_validation_hook(validated)

    @classmethod
    def model_validate_strings(
        cls,
        obj: Any,
        *,
        strict: bool | None = None,
        extra: ExtraValues | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
    ) -> Self:
        return cls.model_validate(
            obj,
            strict=strict,
            extra=extra,
            context=context,
            by_alias=by_alias,
            by_name=by_name,
        )

    @property
    def __annotations__(self) -> dict[str, Any]:  # type: ignore
        return self.get_adapter().__annotations__

    @classmethod
    def read_json_file(cls, read_path: Path) -> Self:
        return cls.model_validate_json(read_path.read_text())

    def write_json_file(self, write_path: Path) -> None:
        write_json(cast(JSONDict, self.model_dump(mode="json")), write_path)

    @staticmethod
    def pre_validation_hook_python(python_dict: dict) -> dict:
        return python_dict

    @staticmethod
    def pre_validation_hook_json(json_string: str | bytes | bytearray) -> str | bytes | bytearray:
        return json_string

    @staticmethod
    def post_validation_hook(validated: T) -> T:
        return validated

    def pre_dump_hook(self, *, mode: Mode) -> Self:
        if mode not in VALID_MODES:
            raise ValueError
        return self

    def post_dump_hook(self, dumped: T, *, mode: Mode) -> T:
        return dumped

    def schema(self) -> CoreSchema:
        return self.get_adapter().core_schema

    def schema_json(self) -> str:
        return json.dumps(self.model_json_schema())

    def _validate_other(self, other) -> Self:
        if not self.__class__ == other.__class__:
            other = self.__class__.model_validate(other)
        return other


class BaseDict[K, V](dict[K, V], AbstractCustom):
    """
    Dictionary type leveraging pydantic for validation and JSON serialization.
    """

    @classmethod
    def get_adapter(cls) -> TypeAdapter[dict[K, V]]:
        """Return a TypeAdapter for this subclass, preserving its key/value types."""
        for base in getattr(cls, "__orig_bases__", []):
            if isinstance(base, GenericAlias):
                key_type, val_type = get_args(base)
                return TypeAdapter(dict[key_type, val_type])  # type: ignore[valid-type]
        raise TypeError("Key and value types not found.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{dict.__repr__(self)}"

    def __str__(self) -> str:
        raise NotImplementedError

    def __or__(self, other: dict[K, V]) -> "BaseDict[K, V]":  # type: ignore
        return self | self._validate_other(other)

    def update(self, other) -> None:
        super().update(self._validate_other(other))


class BaseList[T](list[T], AbstractCustom):
    """
    Dictionary type leveraging pydantic for validation and JSON serialization.
    """

    @classmethod
    def get_adapter(cls) -> TypeAdapter[list[T]]:
        """Return a TypeAdapter for this subclass, preserving its key/value types."""
        for base in getattr(cls, "__orig_bases__", []):
            if isinstance(base, GenericAlias):
                element_type = get_args(base)[0]
                return TypeAdapter(list[element_type])  # type: ignore[valid-type]
        raise TypeError("Key and value types not found.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{list.__repr__(self)}"

    def __str__(self) -> str:
        raise NotImplementedError

    def extend(self, other: Iterable) -> None:
        super().extend(self._validate_other(other))

    def append(self, other: Any) -> None:
        self.extend([other])


class BaseSet[T](set[T], AbstractCustom):
    """
    Dictionary type leveraging pydantic for validation and JSON serialization.

    TODO: make dump as list if not all members hashable.
    """

    @classmethod
    def get_adapter(cls) -> TypeAdapter[set[T]]:
        """Return a TypeAdapter for this subclass, preserving its key/value types."""
        for base in getattr(cls, "__orig_bases__", []):
            if isinstance(base, GenericAlias):
                element_type = get_args(base)[0]
                return TypeAdapter(set[element_type])  # type: ignore[valid-type]
        raise TypeError("Key and value types not found.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{set.__repr__(self)}"

    def __str__(self) -> str:
        raise NotImplementedError

    def __or__(self, other) -> "BaseSet[T]":  # type: ignore
        return self | self._validate_other(other)

    def add(self, other: T) -> None:
        self.update({other})

    def update(self, *s: Iterable[T]) -> None:
        for other in s:
            super().update(self._validate_other(other))

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_computed_fields: bool = False,
        round_trip: bool = False,
        warnings: (bool | Literal["none", "warn", "error"]) = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        if mode not in VALID_MODES:
            raise ValueError
        if exclude_computed_fields:
            raise ValueError

        return BaseList.model_validate(self).model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )



from collections import defaultdict
from collections.abc import Callable
from typing import TypeVar, cast

T = TypeVar("T")
K = TypeVar("K")


class FrozenDefaultDict[K, T](defaultdict[K, T]):
    def __init__(self, default_factory: Callable[[], T], dictionary: dict[K, T]):
        super().__init__(default_factory, dictionary)

    def __repr__(self) -> str:
        default = self.default_factory
        result = default.__name__ if callable(default) else repr(default)
        return f"{self.__class__.__name__}({dict(self)}, default={result})"

    def __getitem__(self, key: K) -> T:
        if key in self:
            return super().__getitem__(key)
        return cast(Callable, self.default_factory)()

    def __setitem__(self, key: K, value: T) -> None:
        raise TypeError(f"{self.__class__.__name__} is immutable")

    def __delitem__(self, key) -> None:
        raise TypeError(f"{self.__class__.__name__} is immutable")



# for tests
sample = [
    {"id": "a", "priority": 0.4, "prerequisites": []},
    {"id": "b", "priority": 0.7, "prerequisites": []},
    {"id": "c", "priority": 0.4, "prerequisites": []},
    {"id": "d", "priority": 0.3, "prerequisites": []},
    {"id": "e", "priority": 0.8, "prerequisites": ["a", "b"]},
    {"id": "f", "priority": 0.9, "prerequisites": []},
    {"id": "g", "priority": 0.1, "prerequisites": ["i", "m"]},
    {"id": "h", "priority": 0.5, "prerequisites": []},
    {"id": "i", "priority": 0.45, "prerequisites": ["j"]},
    {"id": "j", "priority": 0.3, "prerequisites": []},
    {"id": "k", "priority": 0.8, "prerequisites": ["l", "m", "d"]},
    {"id": "l", "priority": 0.6, "prerequisites": ["o"]},
    {"id": "m", "priority": 0.9, "prerequisites": []},
    {"id": "n", "priority": 0.4, "prerequisites": ["o"]},
    {"id": "o", "priority": 0.2, "prerequisites": []},
    {"id": "p", "priority": 0.5, "prerequisites": []},
]


def dep_sort(task_list: list[dict]) -> list[dict]:
    """
    Sort list of dictionaries such that
      1) dependency constraints are satisfied and
      2) priority ordering is satisfied subject to (1)
    """
    # sort on priority to preserve priority order in the output
    task_list.sort(key=lambda t: t["priority"], reverse=True)
    task_ids = [t["id"] for t in task_list]

    # important to make changes in reverse priority order so that insertion preserves priority order
    dep_dict = {t["id"]: t["prerequisites"] for t in reversed(task_list) if t["prerequisites"]}
    print(dep_dict)

    def place_after(to_move: str, deps: list[str], id_list: list) -> bool:
        if not deps:
            return False
        idx1 = id_list.index(to_move)
        idx2 = max(map(id_list.index, deps))
        if idx1 > idx2:
            return False
        id_list.pop(idx1)
        id_list.insert(idx2, to_move)
        return True

    print(task_ids)
    count = 0
    maxcount = sum(range(len(task_ids) + 1))
    while count < maxcount:
        change_tracker = []
        for task_id, task_deps in dep_dict.items():
            changed = place_after(task_id, task_deps, task_ids)
            change_tracker.append(changed)
        unchanged = not any(change_tracker)
        print(task_ids)
        if unchanged:
            return sorted(task_list, key=lambda t: task_ids.index(t["id"]))
        count += 1

    print("ERROR ---------------------------------------------------")
    str(task_ids)
    for task_id, task_deps in dep_dict.items():
        changed = place_after(task_id, task_deps, task_ids)
        if changed:
            after = str(task_ids)
            print(after)

    raise ValueError("Graph contains a cycle.")

    # levels = {t: 0 for t in task_ids}
    # roots = list(filter(lambda t: t["depencies"] == [], task_list))
    # depended_on = {t: set() for t in task_ids}
    # for t in task_list:
    #     for d in t["prerequisites"]:
    #         depended_on[d].add(t)
    # dict_rep = json.dumps(levels)
    # new_rep = ""
    # while new_rep != dict_rep:
    #     dict_rep = json.dumps(levels)
    #     for t in task_list:
    #         levels[t["id"]] = max([levels[d] for d in t["prerequisites"]]) + 1
    #     new_rep = json.dumps(levels)
    # return sorted(task_list, key=lambda t: levels[t["id"]])



from collections import defaultdict
from collections.abc import Callable

# def display_counts(key: str, depth: int, dl: list[str]) -> None:
#     print(f"\n=== {key} ===")
#     for d in dl:
#         if key not in d:
#             print(d["name"])
#     cats = [d[key] for d in dl]
#     cats = [".".join(re.split(r"\.|, ", c)[:depth]) for c in cats]
#     counts = sorted([(cats.count(c), c) for c in sorted(set(cats))], reverse=True)
#     for count, item in counts:
#         print(f"{count:>4} {item}")


def print_tree(strings):
    # Nested dictionary to hold tree structure
    def tree() -> defaultdict:
        return defaultdict(tree)

    root = tree()

    # Build the tree
    for _string in strings:
        parts = _string.split(".")
        current_level = root
        for part in parts:
            current_level = current_level[part]

    # Function to print the tree recursively
    def print_subtree(node, prefix=""):
        children = list(node.keys())
        for i, child in enumerate(children):
            is_last = i == len(children) - 1
            if is_last:
                print(prefix + "└─ " + child)
                new_prefix = prefix + "   "
            else:
                print(prefix + "├─ " + child)
                new_prefix = prefix + "│  "
            print_subtree(node[child], new_prefix)

    # Print the root
    print_subtree(root)


def wrap_line(line: str, length: int, formatter: Callable) -> str:
    raise NotImplementedError



from .color import Colorizer
from .comparison import (
    as_bool,
    equal_within,
    make_maybe_not,
    match_empty,
    nearly_equal,
    trivial_condition,
)
from .exceptions import CustomValidationError
from .file_modification_time import (
    first_newer,
    time_created,
    time_created_readable,
    time_modified,
    time_modified_readable,
)
from .frozendict import FrozenDefaultDict
from .functional import (
    dmap,
    endofilter,
    endomap,
    fold_dictionaries,
    identity,
    kfilter,
    kmap,
    lfilter,
    lmap,
    sfilter,
    smap,
    split_head,
    tail,
    tfilter,
    tmap,
    vfilter,
    vmap,
)
from .import_utils import (
    maybe_import,
    maybe_import_lazy,
)
from .io_utils import (
    ensure_path,
    list_full,
    list_names,
    name_from_full,
    read_json,
    read_raw,
    remove_all_under,
    write_json,
    write_raw,
    write_raw_bytes,
)
from .markers import (
    endo,
    impure,
    mutates,
    mutates_and_returns_instance,
    mutates_instance,
    pure,
    refactor,
)
from .merge import (
    join_as_sequence,
    make_hashable,
    merge_dicts,
)
from .numerical import evenly_spaced, ihash, round5
from .paths_manager import PathsManager
from .performance_logging import log_perf  # type: ignore
from .pydantic_extensions import BaseDict, BaseList, BaseModelRW, BaseSet
from .re_utils import escape_character, re_bool_search, re_split
from .string_utils import (
    MixedValidated,
    PromptTypeName,
    as_json,
    cast_as,
    flexsplit,
    indent_lines,
    parse_sequence,
)
from .timestamping import insert_timestamp, make_timestamp
from .typing_utils import (
    Endofunction,
    JSONDict,
    areinstances,
    call_fallback_if_none,
    fallback_if_none,
)

DELIMITER = "᜶"

__all__ = [
    "DELIMITER",
    "BaseDict",
    "BaseList",
    "BaseModelRW",
    "BaseSet",
    "Colorizer",
    "CustomValidationError",
    "Endofunction",
    "FrozenDefaultDict",
    "JSONDict",
    "MixedValidated",
    "PathsManager",
    "PromptTypeName",
    "areinstances",
    "as_bool",
    "as_json",
    "call_fallback_if_none",
    "cast_as",
    "dmap",
    "endo",
    "endofilter",
    "endomap",
    "ensure_path",
    "equal_within",
    "escape_character",
    "evenly_spaced",
    "fallback_if_none",
    "first_newer",
    "flexsplit",
    "fold_dictionaries",
    "identity",
    "ihash",
    "impure",
    "indent_lines",
    "insert_timestamp",
    "join_as_sequence",
    "kfilter",
    "kmap",
    "lfilter",
    "list_full",
    "list_names",
    "lmap",
    "log_perf",
    "make_hashable",
    "make_maybe_not",
    "make_timestamp",
    "match_empty",
    "maybe_import",
    "maybe_import_lazy",
    "merge_dicts",
    "mutates",
    "mutates_and_returns_instance",
    "mutates_instance",
    "name_from_full",
    "nearly_equal",
    "parse_sequence",
    "pure",
    "re_bool_search",
    "re_split",
    "read_json",
    "read_raw",
    "refactor",
    "remove_all_under",
    "round5",
    "sfilter",
    "smap",
    "split_head",
    "tail",
    "tfilter",
    "time_created",
    "time_created_readable",
    "time_modified",
    "time_modified_readable",
    "tmap",
    "trivial_condition",
    "vfilter",
    "vmap",
    "write_json",
    "write_raw",
    "write_raw_bytes",
]
